@@ -1,11 +1,12 @@
-from PyFWI.wave_propagation import WavePropagator as Wave
 from scipy.optimize.optimize import MemoizeJac
-import PyFWI.optimization as opt
-import PyFWI.fwi_tools as tools
-import PyFWI.acquisition as acq
 import numpy as np
+from scipy.optimize import fmin_cg
 from scipy.optimize.lbfgsb import fmin_l_bfgs_b
+
+from PyFWI.wave_propagation import WavePropagator as Wave
 from PyFWI.fwi_tools import Regularization
+import PyFWI.fwi_tools as tools
+from PyFWI.processing import prepare_residual
 
 class FWI(Wave):
     """
@@ -33,8 +34,8 @@ class FWI(Wave):
     param_functions : dict, optional
         List of functions required in case of inversion with different parameterization than dv, by default None
         """
-    def __init__(self, d_obs, inpa, src, rec_loc, model_shape, n_well_rec, chpr, components, param_functions=None):
-        super().__init__(inpa, src, rec_loc, model_shape, n_well_rec, chpr, components)
+    def __init__(self, d_obs, inpa, src, rec_loc, model_shape, components, chpr, n_well_rec=0, param_functions=None):
+        super().__init__(inpa, src, rec_loc, model_shape, n_well_rec=n_well_rec, chpr=chpr, components=components)
         self.regularization = Regularization(self.nx, self.nz, self.dh, self.dh)
         
         if param_functions is None:
@@ -81,7 +82,7 @@ class FWI(Wave):
         else:
             self.tikhonov_properties = None
             
-        self.d_obs = acq.prepare_residual(d_obs, 1)
+        self.d_obs = prepare_residual(d_obs, 1)
         
         self.fn = inpa['fn']
 
@@ -152,14 +153,41 @@ class FWI(Wave):
         if user_method in [0, 'SD', 'sd']:
             raise ("Steepest descent is not provided yet.")
             # TODO Add option
-        elif user_method in [1, 'GD', 'gd']:
-            raise ("Gradient descent is not provided yet.")
-            # TODO Add option
+        elif user_method in [1, 'CG', 'cg']:
+            method += 'cg'
+            
         elif user_method in [2, 'lbfgs']:
             method += 'lbfgs'
 
         return method
 
+    def cg(self, m0, ITER, freq, n_params=1, k0=1, k_end=2):
+        
+        n_element = self.nz * self.nx
+        mtotal = np.copy(m0)
+
+        rms_hist = []
+
+        fun = MemoizeJac(self.fprime_single)
+        jac = fun.derivative
+        
+        for k in np.arange(k0, k_end, n_params):
+            print('Parameter number {} to {}'.format(k + 1, k + n_params))
+
+            m_1 = mtotal[:k * n_element]
+            m_opt = mtotal[k * n_element: (k + n_params) * n_element]
+            m1 = mtotal[(k + n_params) * n_element:]
+            
+            m_opt, hist, _, _, _ = fmin_cg(fun, m_opt, jac, args=[m_1, m1, freq],
+                                           gtol=1e-8, maxiter=ITER,
+                                           full_output = True, disp=None)
+
+            # print(m_opt.max(), m_opt.min())
+            rms_hist.append(hist)
+
+            mtotal = np.hstack((m_1, m_opt, m1))
+        return mtotal, rms_hist
+    
     def lbfgs(self, m0, ITER, freq, n_params=1, k0=1, k_end=2):
         
         n_element = self.nz * self.nx
@@ -195,7 +223,7 @@ class FWI(Wave):
         m_new = self.model_to_dv(m_old, self.param_functions_args)
 
         d_est = self.forward_modeling(m_new, show=False)
-        d_est = acq.prepare_residual(d_est, self.sd)
+        d_est = prepare_residual(d_est, self.sd)
 
         rms_data, adj_src = tools.cost_seismic(d_est, self.d_obs, fun=self.CF,
                                                fn=self.fn, freq=freq, order=3, axis=1
@@ -214,7 +242,7 @@ class FWI(Wave):
         return rms, grad
     
     def fprime_single(self, m0, m_1, m1, freq):
-        mtotal = np.hstack((m_1, m0, m1))
+        mtotal = np.float32(np.hstack((m_1, m0, m1)))
         shape_1 = np.shape(m_1)[0]
         shape0 = np.shape(m0)[0]
         
