@@ -1,14 +1,15 @@
-from PyFWI.wave_propagation import WavePropagator as Wave
 from scipy.optimize.optimize import MemoizeJac
-import PyFWI.fwi_tools as tools
-import numpy as np
 from scipy.optimize.lbfgsb import fmin_l_bfgs_b
+import numpy as np
 import matplotlib.pyplot as plt
 from scipy.optimize import line_search
 import copy
-from PyFWI.fwi import FWI
-import PyFWI.acquisition as acq
 import timeit
+import PyFWI.fwi_tools as tools
+from PyFWI.fwi import FWI
+from PyFWI.processing import prepare_residual
+from PyFWI.wave_propagation import WavePropagator as Wave
+
 
 
 class TimeLapse(Wave):
@@ -38,10 +39,11 @@ class TimeLapse(Wave):
     param_functions : dict, optional
         Required functions to switch the gradient
         """
-    def __init__(self, b_dobs, m_dobs, inpa, src, rec_loc, model_size, n_well_rec, chpr, components, param_functions=None):
-        super().__init__(inpa, src, rec_loc, model_size, n_well_rec, chpr, components)
+    def __init__(self, b_dobs, m_dobs, inpa, src, rec_loc, model_shape, components, chpr, n_well_rec=0, param_functions=None):
+        super().__init__(inpa, src, rec_loc, model_shape, n_well_rec=n_well_rec, chpr=chpr, components=components)
         
-        self.inv_obj = FWI(b_dobs, inpa, src, rec_loc, model_size, n_well_rec, chpr, components, param_functions)
+        self.inv_obj = FWI(b_dobs, inpa, src, rec_loc, model_shape=model_shape,
+                           n_well_rec=n_well_rec, chpr=chpr, components=components, param_functions=param_functions)
 
         self.b_dobs = copy.deepcopy(b_dobs)
         self.m_dobs = copy.deepcopy(m_dobs)
@@ -83,9 +85,13 @@ class TimeLapse(Wave):
             rms error
         """
                 
-        # b_m0 must remain a dict    
-        if tl_method == 'cc':
+        # b_m0 must remain a dict
+        if tl_method == 'pa':
+            m, rms = self.parallel_inversion(b_m0, freqs, iter, n_params=n_params, k_0=k_0, k_end=k_end)    
+        elif tl_method == 'cc':
             m, rms = self.cascaded_inversion(b_m0, freqs, iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        elif tl_method == 'dd':
+            m, rms = self.double_difference_inversion(b_m0, freqs, iter, n_params, k_0, k_end)
         elif tl_method == 'sim':
             m, rms = self.sim_inversion(b_m0, freqs, iter, data_to_invert=self.b_dobs, n_params=n_params, k_0=k_0, k_end=k_end)
         elif tl_method == 'wa':
@@ -157,7 +163,7 @@ class TimeLapse(Wave):
         m_model0 = self.inv_obj.model_to_dv(m_model0_pcs, self.inv_obj.param_functions_args)
         
         dpre = self.forward_modeling(b_model0, show=False)
-        dpre = acq.prepare_residual(dpre, self.inv_obj.sd)
+        dpre = prepare_residual(dpre, self.inv_obj.sd)
         
         b_rms_data, adj_src1 = tools.cost_seismic(d_pre0=dpre, d_obs0=self.b_dobs, fun=self.CF,
                                          fn=self.fn, freq=freq, order=3, axis=1,
@@ -180,7 +186,7 @@ class TimeLapse(Wave):
         b_grad = self.inv_obj.grad_from_dv(b_grad_dv, self.inv_obj.param_functions_args, b_model0_pcs)
         
         dpre = self.forward_modeling(m_model0, show=False)
-        dpre = acq.prepare_residual(dpre, self.inv_obj.sd)
+        dpre = prepare_residual(dpre, self.inv_obj.sd)
         
         m_rms_data, adj_src = tools.cost_seismic(d_pre0=dpre, d_obs0=self.m_dobs, fun=self.CF,
                                            fn=self.fn, freq=freq, order=3, axis=1,
@@ -214,15 +220,15 @@ class TimeLapse(Wave):
         
         print("*** First inversion ***")
         self.inv_obj.d_obs = self.b_dobs
-        model_b1, rms1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b1, rms_b1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Second inversion ***")
         self.inv_obj.d_obs = self.m_dobs
-        model_m1, rms2 = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_m1, rms_m1 = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Third inversion ***")
         self.inv_obj.d_obs = self.b_dobs
-        model_b2, rms3 = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b2, rms_b2 = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
         
         inverted = {
             "dm11": tools.dict_diff(model_m1, model_b1),
@@ -231,8 +237,15 @@ class TimeLapse(Wave):
             "b1": model_b1,
             "b2": model_b2,
         }
-        print("improved_cascaded")
-        return inverted, np.array(rms1 + rms2 + rms3) / 3
+        print("Weighted average")
+        
+        hist = {
+            "b1": rms_b1,
+            "m1": rms_m1,
+            "b2": rms_b2
+        }
+        
+        return inverted, hist
     
     def cross_updating(self, b_m0, freqs, iter, n_params, k_0, k_end):
         """
@@ -242,21 +255,21 @@ class TimeLapse(Wave):
         print("*** First inversion ***")
         tic = timeit.default_timer()
         self.inv_obj.d_obs = self.b_dobs
-        model_b1, rms1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b1, rms_b1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Second inversion ***")
         self.inv_obj.d_obs = self.m_dobs
-        model_m1, rms2 = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_m1, rms_m1 = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
         process_time_cc = timeit.default_timer() - tic
         
         print("*** Third inversion ***")
         self.inv_obj.d_obs = self.b_dobs
-        model_b2, rms3 = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b2, rms_b2 = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
         process_time_wa = timeit.default_timer() - tic
         
         print("*** Fourth inversion ***")
         self.inv_obj.d_obs = self.m_dobs
-        model_m2, rms4 = self.inv_obj(model_b2, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_m2, rms_m2 = self.inv_obj(model_b2, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
         process_time_cu = timeit.default_timer() - tic
 
         inverted = {
@@ -273,29 +286,39 @@ class TimeLapse(Wave):
             "pt_cu": process_time_cu
         }
         print("cross_updating")
-        return inverted, np.array(rms1 + rms2 + rms3 + rms4) / 4
+        
+        hist = {
+            "b1": rms_b1,
+            "m1": rms_m1,
+            "b2": rms_b2,
+            "m2": rms_m2
+        }
+        
+        return inverted, hist
     
     
     def central_difference(self, b_m0, freqs, iter, n_params, k_0, k_end):
         
         m0 = copy.deepcopy(b_m0)
             # ============ Central difference for parallel ==============
-
+        tic = timeit.default_timer()        
         print("*** Central time lapse *** ")
         self.inv_obj.d_obs = self.b_dobs
-        model_b1, rms1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b1, rms_b1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Second INVERSION ***")
         self.inv_obj.d_obs = self.m_dobs
-        model_b1m, _ = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_b1m, rms_b1m = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Third INVERSION ***")
         self.inv_obj.d_obs = self.m_dobs
-        model_m1, _ = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_m1, rms_m1 = self.inv_obj(m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
         print("*** Fourth INVERSION ***")
         self.inv_obj.d_obs = self.b_dobs
-        model_m1b, _ = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        model_m1b, rms_m1b = self.inv_obj(model_m1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        
+        process_time = timeit.default_timer() - tic
 
         inverted = {
             "dm0": tools.dict_diff(model_m1,  model_b1),
@@ -308,15 +331,25 @@ class TimeLapse(Wave):
             "b1": model_b1,
             "m1": model_m1,
             "m2": model_b1m,
-            "b2": model_m1b
+            "b2": model_m1b,
+            'pt': process_time
         }
-        print("central_inversion")
-        return inverted, _
+        print("central difference")
+        
+        hist = {
+            "b1": rms_b1,
+            "m1": rms_m1,
+            "m2": rms_b1m,
+            "b2": rms_m1b
+        }
+        
+        return inverted, hist
         
     def cascaded_joint(self, m0, freqs, iter, n_params, k_0, k_end):
         """
             
         """
+        tic = timeit.default_timer()
         # ============ Central difference for sim ==============
         print("*** FIRST INVERSION ***")
         model1, rms1 = self.sim_inversion(m0, freqs, iter=iter,n_params=n_params,
@@ -331,7 +364,9 @@ class TimeLapse(Wave):
                                        k_0=k_0, k_end=k_end, data_to_invert=None)
         model_b1b = model_b1x["b1"]
         model_b1m = model_b1x["m1"]
-
+        
+        process_time = timeit.default_timer() - tic
+        
         inverted = {
             "dm0": tools.dict_diff(model_m1,  model_b1),
             "dm7": tools.dict_diff(model_b1m, model_b1b),
@@ -339,48 +374,133 @@ class TimeLapse(Wave):
             "dm3": tools.dict_diff(model_b1m, model_b1),
             "dm6": tools.dict_diff(model_m1, model_b1b),
 
-            "model_b1": model_b1,
-            "model_m1": model_m1,
+            "b1": model_b1,
+            "m1": model_m1,
 
-            "model_b2": model_b1b,
-            "model_m2": model_b1m,
+            "b2": model_b1b,
+            "m2": model_b1m,
+            
+            "pt": process_time
+        }
+        
+        print("joint_cascaded")
+        
+        hist = {
+            "sim1": rms1,
+            "sim2": rms2
+        }
+        
+        return inverted, hist
+    
+    def parallel_inversion(self, b_m0, freqs, iter, n_params, k_0, k_end):
+        
+        tic = timeit.default_timer()
+        self.inv_obj.d_obs = self.b_dobs
+        model_b1, rms1 = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+
+        print('===== Inversion of baseline is done ======')
+        
+        self.inv_obj.d_obs = self.m_dobs
+        model_m1, rms2 = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        process_time = timeit.default_timer() - tic
+        
+        inverted = {
+            "b1": model_b1,
+            "m1": model_m1,
+            "dm":  tools.dict_diff(model_m1, model_b1, positivity=False),
+            "pt": process_time
+        }
+        
+        print("Parallel")
+
+        hist = {
+            "b1": rms1,
+            "m1": rms2
         }
 
-        print("joint_cascaded")
-        return inverted, (rms1 + rms2) / 2
+        return inverted, hist
+    
+    def double_difference_inversion(self, b_m0, freqs, iter, n_params, k_0, k_end):
+        """
+            Based on Asnaashari et al., 2011, Sensitivity analysis of time-lapse images obtained
+            # by differential waveform inversion with respect to reference model
+        """
+        tic = timeit.default_timer()
+        
+        self.inv_obj.d_obs = self.b_dobs
+        model_b1, rms1 = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        print(model_b1['vp'].min(), model_b1['vp'].mean(), model_b1['vp'].max())
+
+        print('===== Inversion of baseline is done ======')
+        d_ref = self.forward_modeling(model_b1, show=False)
+        d_ref = prepare_residual(d_ref, s=self.inv_obj.sd)
+        
+        d_composit = tools.dict_summation(tools.dict_diff(self.m_dobs, self.b_dobs), d_ref)
+
+        self.inv_obj.d_obs = d_composit
+
+        model_ref1 = b_m0.copy()
+        model_composite, rms2 = self.inv_obj(model_ref1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+        print(model_b1['vp'].min(), model_b1['vp'].mean(), model_b1['vp'].max())
+
+        model_dm = tools.dict_diff(model_composite, model_b1, positivity=False)
+        process_time = timeit.default_timer() - tic
+        
+        inverted = {
+            "b1": model_b1,
+            "m1": model_composite,
+            "dm": model_dm,
+            "pt": process_time
+        }
+        
+        print("double_difference_inversion")
+        
+        hist = {
+            "b1": rms1,
+            "m1": rms2
+        }
+        return inverted, hist
     
     def cascaded_inversion(self, b_m0, freqs, iter, n_params, k_0, k_end):
 
+        tic = timeit.default_timer()
         self.inv_obj.d_obs = self.b_dobs
         model_b1, rms1 = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
-        initial = model_b1.copy()
+        initial = model_b1.copy() #TODO: WHY?
 
         print('===== Inversion of baseline is done ======')
         
         self.inv_obj.d_obs = self.m_dobs
         model_m1, rms2 = self.inv_obj(model_b1, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
+        process_time = timeit.default_timer() - tic
         inverted = {
             "b1": model_b1,
             "m1": model_m1,
-            "dm":  tools.dict_diff(model_m1, initial, positivity=False)
+            "dm":  tools.dict_diff(model_m1, initial, positivity=False),
+            "pt": process_time
         }
         print("Cascaded")
 
-        hist = rms1 + rms2
-
-        return inverted, np.array(hist)
+        hist = {
+            "b1": rms1,
+            "m1": rms2
+        }
+        
+        return inverted, hist
     
     def sim_inversion(self, b_m0, freqs, iter, n_params, k_0, k_end, data_to_invert=None):
 
-        hist = []  
-
+        rms2 = []  
+        tic = timeit.default_timer()
+        hist = {}
         if data_to_invert is not None:
             self.dobs = data_to_invert
-            model_ref0, _ = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
+            model_ref0, rms1 = self.inv_obj(b_m0, method=2, freqs=freqs, iter=iter, n_params=n_params, k_0=k_0, k_end=k_end)
 
             b_model0 = copy.deepcopy(model_ref0)
             m_model0 = copy.deepcopy(model_ref0)
+            hist["b1"] = rms1
 
         else:
             b_model0 = copy.deepcopy(b_m0)
@@ -393,24 +513,29 @@ class TimeLapse(Wave):
 
         # method = self.optimization_method(operation="time-lapse")
 
-        iter_num = 0
         c = 0
         for freq in freqs:
             self.restForTest = x0
 
             x0, hist1 = self.tl_lbfgs(x0, iter[c], freq, n_params, k_0, k_end)
 
-            hist.append(hist1[0])
+            rms2.append(hist1[0])
             c += 1
         
         models = self.ins_TL_output(x0, False)
+        process_time = timeit.default_timer() - tic
         inverted = {
             "dm": self.ins_TL_output(x0, True),
             "b1": models["model_b"],
-            "m1": models["model_m"]
+            "m1": models["model_m"],
+            "pt": process_time
+        }
+        
+        hist = {
+            "sim": rms2
         }
 
-        return inverted, np.array(hist/max(hist))
+        return inverted, hist
     
     
     def ins_TL_output(self, xk, diff=True):
