@@ -1,23 +1,21 @@
-import numpy as np
-import sys
-from numpy.core.shape_base import block
-import pyopencl as cl
 import os
+import numpy as np
+# from numpy.core.shape_base import block
+import pyopencl as cl
 from pyopencl.tools import get_test_platforms_and_devices
 import matplotlib.pyplot as plt
-import copy
 from scipy.ndimage import gaussian_filter
 
-import PyFWI.processing as seis_process
 import PyFWI.fwi_tools as tools
-from PyFWI.fwi_tools import Recorder, expand_model, CPML
+from PyFWI.fwi_tools import expand_model
 import PyFWI.acquisition as acq
+from PyFWI.processing import prepare_residual
 from PyFWI.grad_swithcher import grad_lmd_to_vd
 
 
 class WavePreparation:
 
-    def __init__(self, inpa, src, rec_loc, model_shape, n_well_rec=0, chpr=10, components=0, set_env_variable=True):
+    def __init__(self, inpa, src, rec_loc, model_shape, components=0, n_well_rec=0, chpr=10, set_env_variable=True):
         '''
         A class to prepare the variable and basic functions for wave propagation.
 
@@ -34,38 +32,61 @@ class WavePreparation:
         self.nx = np.int32(model_shape[1])
         self.nz = np.int32(model_shape[0])
 
+        if 'seimogram_shape' in keys:
+            self.seismo_shape  = inpa['seimogram_shape']
+        else: # if is not defined, make seismogram as [nt, nr * ns]
+            self.seismo_shape = '2d'
+            
         if 'g_smooth' in keys:
             self.g_smooth = inpa['g_smooth']
         else:
             self.g_smooth = 0
 
-
         if 'npml' in keys:
             self.npml = inpa['npml']
+            pmlR = inpa['pmlR']
+            pml_dir = inpa['pml_dir']
         else:
             self.npml = 0
-
-        # Number of samples in x- and z- direction by considering pml
-        self.tnx = np.int32(self.nx + 2 * self.npml)
-        self.tnz = np.int32(self.nz + 2 * self.npml)
-
-        self.dh = inpa['dh']
-
+            pmlR = 0
+            pml_dir = 2
+            
         if 'sdo' in keys:
             self.sdo = np.int32(inpa['sdo'] / 2)
         else:
             self.sdo = 2
 
-        self.srcx = np.int32(src.i + inpa['npml'])
-        self.srcz = np.int32(src.j + inpa['npml'])
+        if 'rec_dis' in keys:
+            rec_dis = inpa['rec_dis']
+        else:
+            rec_dis = (rec_loc[1,:] - rec_loc[0,:]).max()
+        
+        if 'acq_type' in keys:
+            self.acq_type = inpa['acq_type']
+        else:
+            self.acq_type = 1
+            
+        if 'energy_balancing' in keys:
+            self.energy_balancing = inpa['energy_balancing']
+        else:
+            self.energy_balancing = False
+            
+            
+        # Number of samples in x- and z- direction by considering pml
+        self.tnx = np.int32(self.nx + 2 * self.npml)
+        self.tnz = np.int32(self.nz + 2 * self.npml)
+
+        self.dh = np.float32(inpa['dh'])
+
+        self.srcx = np.int32(src.i + self.npml)
+        self.srcz = np.int32(src.j + self.npml)
         src_loc = np.vstack((self.srcx, self.srcz)).T
 
         self.src = src
         self.ns = np.int32(src.i.size)
+        
+        self.dxr = np.int32(rec_dis / self.dh)
 
-        self.dxr = np.int32(inpa['rec_dis'] / self.dh)
-
-        self.chpr = chpr
         chp = int(chpr * self.nt / 100)
         self.chp = np.linspace(0, self.nt-1, chp, dtype=np.int32)
         if (len(self.chp) < 2): #& (chpr != 0)
@@ -79,14 +100,13 @@ class WavePreparation:
         self.n_well_rec = n_well_rec
         self.n_surface_rec = self.nr - 2 * n_well_rec
 
-        self.acq_type = inpa["acq_type"]
         if n_well_rec ==0 and self.acq_type == 2:
             raise Exception(" Number of geophons in the wells is not defined")
 
         # The matrix containg the geometry of acquisittion (Never used really)
         data_guide = acq.acquisition_plan(self.ns, self.nr, src_loc, self.rec_loc, self.acq_type, n_well_rec, self.dh)
 
-        self.data_guide_sampling = acq.discretized_acquisition_plan(data_guide, self.dh, inpa['npml'])
+        self.data_guide_sampling = acq.discretized_acquisition_plan(data_guide, self.dh, self.npml)
 
         self.rec_top_left_const = np.int32(0)
         self.rec_top_left_var = np.int32(0)
@@ -95,37 +115,32 @@ class WavePreparation:
         self.rec_surface_const = np.int32(0)
         self.rec_surface_var = np.int32(0)
 
-        if inpa["acq_type"] == 0:
-            self.rec_top_right_const = np.int32(rec_loc[0, 0] / self.dh + inpa['npml'])
-            self.rec_top_right_var = np.int32(rec_loc[:, 1] / self.dh + inpa['npml'])[0]
+        if self.acq_type == 0:
+            self.rec_top_right_const = np.int32(rec_loc[0, 0] / self.dh + self.npml)
+            self.rec_top_right_var = np.int32(rec_loc[0, 1] / self.dh + self.npml)
             self.src_cts = src.i[0]
 
-        elif inpa["acq_type"] == 1:
-            self.rec_surface_const = np.int32(rec_loc[0, 1] / self.dh + inpa['npml'])
-            self.rec_surface_var = np.int32(rec_loc[:, 0] / self.dh + inpa['npml'])[0]
+        elif self.acq_type == 1:
+            self.rec_surface_const = np.int32(rec_loc[0, 1] / self.dh + self.npml)
+            self.rec_surface_var = np.int32(rec_loc[0, 0] / self.dh + self.npml)
             self.src_cts = src.j[0]
 
-        elif inpa["acq_type"] == 2:
-            a = np.int32(rec_loc[-self.n_well_rec, :]/ self.dh + inpa['npml'])
+        elif self.acq_type == 2:
+            a = np.int32(rec_loc[-self.n_well_rec, :]/ self.dh + self.npml)
             self.rec_top_right_const = np.copy(a[0])
             self.rec_top_right_var = np.copy(a[1])
 
-            a = np.int32(rec_loc[self.n_well_rec, :]/ self.dh + inpa['npml'])
+            a = np.int32(rec_loc[self.n_well_rec, :]/ self.dh + self.npml)
             self.rec_surface_const = np.copy(a[1])
             self.rec_surface_var = np.copy(a[0])
 
-            self.rec_top_left_const = np.int32(rec_loc[0, 0] / self.dh + inpa['npml'])
-            self.rec_top_left_var = np.int32(rec_loc[:, 1] / self.dh + inpa['npml'])[0]
+            self.rec_top_left_const = np.int32(rec_loc[0, 0] / self.dh + self.npml)
+            self.rec_top_left_var = np.int32(rec_loc[:, 1] / self.dh + self.npml)[0]
             self.src_cts = src.j[0]
 
         # ======== Parameters Boundary condition ======
         self.dx_pml, self.dz_pml = tools.pml_counstruction(self.tnz, self.tnx, self.dh, self.npml,
-                                                     inpa['pmlR'], inpa['pml_dir'])
-
-        if 'energy_balancing' in keys:
-            self.energy_balancing = inpa['energy_balancing']
-        else:
-            self.energy_balancing = False
+                                                     pmlR, pml_dir)
 
         self.W = {
             'vx': np.zeros((self.tnz, self.tnx, self.ns, self.nchp), dtype=np.float32),
@@ -138,7 +153,6 @@ class WavePreparation:
         self.D = tools.Fdm(order=self.sdo * 2)
 
         self.components = components
-        self.R = Recorder(self.nt, self.rec_loc, self.ns, self.dh)
 
         # To call openCl
         # Select the platform (if not provided, pick 0)
@@ -170,10 +184,10 @@ class WavePreparation:
         self.prg = cl.Program(self.ctx, kernel).build()
         self.prg_surf = cl.Program(self.ctx, kernel_surface).build()
 
-        if inpa["acq_type"] == 0:
+        if self.acq_type == 0:
             self.prg.injSrc = self.prg_cw.injSrc
             self.prg.Adj_injSrc = self.prg_cw.Adj_injSrc
-        elif inpa["acq_type"] == 1:
+        elif self.acq_type == 1:
             self.prg.injSrc = self.prg_surf.injSrc
             self.prg.Adj_injSrc = self.prg_surf.Adj_injSrc
 
@@ -224,11 +238,11 @@ class WavePreparation:
 
         # Make a list for seismograms
         self.seismogram = {
-            'vx': np.zeros((self.nt, self.nr * self.ns)).astype(np.float32, order='C'),
-            'vz': np.zeros((self.nt, self.nr * self.ns)).astype(np.float32, order='C'),
-            'taux': np.zeros((self.nt, self.nr * self.ns)).astype(np.float32, order='C'),
-            'tauz': np.zeros((self.nt, self.nr * self.ns)).astype(np.float32, order='C'),
-            'tauxz': np.zeros((self.nt, self.nr * self.ns)).astype(np.float32, order='C'),
+            'vx': np.zeros((self.nt, self.nr, self.ns)).astype(np.float32, order='C'),
+            'vz': np.zeros((self.nt, self.nr, self.ns)).astype(np.float32, order='C'),
+            'taux': np.zeros((self.nt, self.nr, self.ns)).astype(np.float32, order='C'),
+            'tauz': np.zeros((self.nt, self.nr, self.ns)).astype(np.float32, order='C'),
+            'tauxz': np.zeros((self.nt, self.nr, self.ns)).astype(np.float32, order='C'),
         }
 
     def adjoint_buffer_preparing(self):
@@ -336,19 +350,19 @@ class WavePreparation:
             cl.enqueue_copy(self.queue, seismogram_id, buffer)
             return np.copy(seismogram_id)
 
-        self.seismogram['vx'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr] = \
+        self.seismogram['vx'][np.int32(t - 1), :, s] = \
             get_from_opencl(self.seismogramid_vx_b)
 
-        self.seismogram['vz'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr] = \
+        self.seismogram['vz'][np.int32(t - 1), :, s] = \
             get_from_opencl(self.seismogramid_vz_b)
 
-        self.seismogram['taux'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr] = \
+        self.seismogram['taux'][np.int32(t - 1), :, s] = \
             get_from_opencl(self.seismogramid_taux_b)
 
-        self.seismogram['tauz'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr] = \
+        self.seismogram['tauz'][np.int32(t - 1), :, s] = \
             get_from_opencl(self.seismogramid_tauz_b)
 
-        self.seismogram['tauxz'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr] = \
+        self.seismogram['tauxz'][np.int32(t - 1), :, s] = \
             get_from_opencl(self.seismogramid_tauxz_b)
 
     def make_residual(self, res, s, t):
@@ -383,7 +397,6 @@ class WavePreparation:
         res_src_tauxz = (res['tauxz'][np.int32(t - 1), s * self.nr:(s + 1) * self.nr]).astype(np.float32, order='C')
         cl.enqueue_copy(self.queue, self.res_tauxz_b, res_src_tauxz)
 
-
     def pml_preparation(self, v_max):
 
         vdx_pml = self.dx_pml * v_max
@@ -393,7 +406,6 @@ class WavePreparation:
         self.vdz_pml_b = cl.Buffer(self.ctx, self.mf.READ_WRITE |
                                    self.mf.COPY_HOST_PTR, hostbuf=vdz_pml)
 
-
     def kernel_caller(self):
         '''
             This function is used to specify the constants for use in openCl's
@@ -601,9 +613,10 @@ class WavePropagator(WavePreparation):
     component:
         Seismic output
     """
-    def __init__(self, inpa, src, rec_loc, model_shape, n_well_rec=None, chpr=10, components=0, set_env_variable=True):
-        WavePreparation.__init__(self, inpa, src, rec_loc, model_shape, n_well_rec, chpr=chpr,
-                                 components=components, set_env_variable=set_env_variable)
+    def __init__(self, inpa, src, rec_loc, model_shape, components=0, n_well_rec=0, chpr=0, set_env_variable=True):
+        WavePreparation.__init__(self, inpa=inpa, src=src, rec_loc=rec_loc, model_shape=model_shape, components=components, 
+                                 n_well_rec=n_well_rec, chpr=chpr,
+                                 set_env_variable=set_env_variable)
 
     def forward_propagator(self, model):
         """ This function is in charge of forward modelling for acoustic case
@@ -645,6 +658,7 @@ class WavePropagator(WavePreparation):
                             self.seismogramid_taux_b, self.seismogramid_tauz_b, self.seismogramid_tauxz_b,
                             self.dxr,
                             self.srcx[s], self.srcz[s],
+                            src_kv_x, src_kv_z,
                             src_kt_x, src_kt_z)
 
             self.prg.update_velx(self.queue, (self.tnz, self.tnx), None,
@@ -852,7 +866,7 @@ class WavePropagator(WavePreparation):
         self.pml_preparation(model['vp'].max())
         self.elastic_buffers(model)
         seismo = self.forward_propagator(model)
-        data = acq.seismic_section(seismo, self.components)
+        data = acq.seismic_section(seismo, self.components, shape=self.seismo_shape)
         return data
 
     def gradient(self, res, show=False, parameterization='dv'):
@@ -881,7 +895,7 @@ class WavePropagator(WavePreparation):
             for par in self.seismogram:
                 self.seismogram[par][:, :self.n_well_rec] = np.flip(self.seismogram[par][:, :self.n_well_rec], axis=1)
 
-        res = acq.prepare_residual(res, 1)
+        res = prepare_residual(res, 1)
         if show:
             self.initial_wavefield_plot({'vp':self.vp}, plot_type="Backward")
 
@@ -911,7 +925,8 @@ if __name__ == "__main__":
     import PyFWI.model_dataset as md
     import PyFWI.acquisition as acq
     import PyFWI.seiplot as splt
-
+    import PyFWI.fwi as fwi
+    
     GRADIENT = 1
     INVERSION = 0
 
@@ -961,7 +976,7 @@ if __name__ == "__main__":
 
     W = WavePropagator(inpa, src, rec_loc, model_shape, n_well_rec, chpr=0, components=inpa['seisout'])
     d_obs = W.forward_modeling(model, False)
-    d_obs = acq.prepare_residual(d_obs, 1)
+    d_obs = prepare_residual(d_obs, 1)
 
     fig = plt.figure()
     ax = fig.add_subplot(1, 1, 1)
@@ -972,7 +987,7 @@ if __name__ == "__main__":
     if GRADIENT:
         Lam = WavePropagator(inpa, src, rec_loc, model_shape, n_well_rec, chpr=chpr, components=inpa['seisout'])
         d_est = Lam.forward_modeling(m0, show=False)
-        d_est = acq.prepare_residual(d_est, 1)
+        d_est = prepare_residual(d_est, 1)
 
         CF = tools.CostFunction('l2')
         rms, res = CF(d_est, d_obs)
@@ -986,7 +1001,6 @@ if __name__ == "__main__":
         a=1
 
     elif INVERSION:
-        import PyFWI.fwi as fwi
 
         FWI = fwi.FWI(d_obs, inpa, src, rec_loc, model_shape, n_well_rec, chpr=chpr, components=4, param_functions=None)
         inverted_model, rms = FWI(m0, method=2, iter=[3], freqs=[25], n_params=1, k_0=1, k_end=2)
